

module map_28
(map_out, cpu_dat, cpu_addr, cpu_ce, cpu_rw, m2,
ppu_dat, ppu_addr, ppu_oe, ppu_we, 
clk, map_cfg, rst, map_idx);
 
	
	output [48:0]map_out;
 
	//cpu bus
	input [7:0]cpu_dat;
	input [14:0]cpu_addr;
	input cpu_ce, cpu_rw, m2;
	
	//ppu bus
	input [7:0]ppu_dat;
	input [13:0]ppu_addr;
	input ppu_oe, ppu_we;
	
	input clk, rst;
	input [7:0]map_cfg;
	input [7:0]map_idx;
	
	//mapper outputs
	wire ciram_a10, ciram_ce;
	wire [19:0]prg_addr;
	wire [19:0]chr_addr;
	wire prg_we;
	wire chr_we;
	wire irq;
	wire rom_ce;
	wire ram_ce;
	wire snd = 1'bzz;
	
	assign map_out[48:0] = {snd, chr_ce, ram_ce, rom_ce, irq, chr_we, prg_we, ciram_ce, ciram_a10, chr_addr[19:0], prg_addr[19:0]};
	//*************************************************************

	assign prg_we = !cpu_rw & ram_ce;
	assign ram_ce = cpu_addr[14:13] == 2'b11 & cpu_ce & m2;
	assign rom_ce = !cpu_ce;
	assign chr_ce = ciram_ce;
	assign chr_we = !ppu_we & ciram_ce;
	
	//A10-Vmir, A11-Hmir
	assign ciram_a10 = !mirror_mode[1] ? mirror_mode[0] : !mirror_mode[0] ? ppu_addr[10] : ppu_addr[11];
	assign ciram_ce = !ppu_addr[13];
	
	assign prg_addr[12:0] = cpu_addr[12:0];
	assign prg_addr[13] = cpu_ce ? 0 : cpu_addr[13];
	assign prg_addr[18:14] = cpu_ce ? 0 : prg[4:0];
	
	assign chr_addr[12:0] = ppu_addr[12:0];
	assign chr_addr[14:13] = chr_ram_bank[1:0];
	

	reg [1:0]reg_addr;
	
	reg [1:0]chr_ram_bank;
	reg [1:0]mirror_mode;
	
	reg [3:0]ine_prg;
	reg [5:0]out_prg;
	reg [1:0]prg_mode;
	reg [1:0]prg_size;
	
	wire [2:0]size_mask = 
	prg_size[1:0] == 0 ? 3'b000 : 
	prg_size[1:0] == 1 ? 3'b001 : 
	prg_size[1:0] == 2 ? 3'b011 : 3'b111;
	
	wire fixed_bank = !prg_mode[1] ? 0 : !prg_mode[0] ? !cpu_addr[14] : cpu_addr[14];
	wire [4:0]prg;
	assign prg[0] = !prg_mode[1] ? cpu_addr[14] : fixed_bank ? cpu_addr[14] : ine_prg[0];
	assign prg[1] = size_mask[0] & !fixed_bank  ? ine_prg[1] : out_prg[0];
	assign prg[2] = size_mask[1] & !fixed_bank  ? ine_prg[2] : out_prg[1];
	assign prg[3] = size_mask[2] & !fixed_bank  ? ine_prg[3] : out_prg[2];
	assign prg[4] = out_prg[3];
	
	//assign prg_addr[18:15] = 
	
	 //wire [4:0]mode_a = { cpu_addr[14]}
	 
	
	always @(negedge m2)
	begin
	
		if(rst)
		begin
			ine_prg[3:0] <= 0;
			out_prg[5:0] <= 5'b11111;
			mirror_mode[1:0] <= 2'b01;
		end
			else
		if(!cpu_rw & cpu_ce & cpu_addr[14:12] == 3'b101)reg_addr[1:0] <= {cpu_dat[7], cpu_dat[0]};
			else
		if(!cpu_rw & !cpu_ce)
		case(reg_addr)
			0:begin
				chr_ram_bank[1:0] <= cpu_dat[1:0];
				if(!mirror_mode[1])mirror_mode[0] <= cpu_dat[4];
			end
			1:begin
				ine_prg[3:0] <= cpu_dat[3:0];
				if(!mirror_mode[1])mirror_mode[0] <= cpu_dat[4];
			end
			2:begin
				mirror_mode[1:0] <= cpu_dat[1:0];
				prg_mode[1:0] <= cpu_dat[3:2];
				prg_size[1:0] <= cpu_dat[5:4];
			end
			3:begin
				out_prg[5:0] <= cpu_dat[5:0];
			end
		endcase
		
	end
	
	
endmodule
